@@ -1,24 +1,24 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2014 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
  *   Free Software Foundation; either version 2 of the License, or (at your
  *   option) any later version.
- * 
+ *
  *   This program is distributed in the hope that it will be useful,
  *   but WITHOUT ANY WARRANTY; without even the implied warranty of
  *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  *   See the GNU General Public License for more details.
- * 
+ *
  *   You should have received a copy of the GNU General Public License along
  *   with this program; if not, see <http://www.gnu.org/licenses>.
- * 
+ *
  *   Linking stunnel statically or dynamically with other modules is making
  *   a combined work based on stunnel. Thus, the terms and conditions of
  *   the GNU General Public License cover the whole combination.
- * 
+ *
  *   In addition, as a special exception, the copyright holder of stunnel
  *   gives you permission to combine stunnel with free software programs or
  *   libraries that are released under the GNU LGPL and with code included
@@ -26,7 +26,7 @@
  *   modified versions of such code, with unchanged license). You may copy
  *   and distribute such a system following the terms of the GNU GPL for
  *   stunnel and the licenses of the other code concerned.
- * 
+ *
  *   Note that people who make modified versions of stunnel are not obligated
  *   to grant this special exception for their modified versions; it is their
  *   choice whether to do so. The GNU General Public License gives permission
@@ -42,45 +42,50 @@
 
 #include <tcpd.h>
 
-static int check(char *, int);
+#if defined(USE_PTHREAD) && !defined(__CYGWIN__)
+/* http://wiki.osdev.org/Cygwin_Issues#Passing_file_descriptors */
+#define USE_LIBWRAP_POOL
+#endif /* USE_PTHREAD && !__CYGWIN__ */
+
+NOEXPORT int check(char *, int);
 
 int allow_severity=LOG_NOTICE, deny_severity=LOG_WARNING;
 
-#ifdef USE_PTHREAD
+#ifdef USE_LIBWRAP_POOL
 #define SERVNAME_LEN 256
 
-static ssize_t read_fd(int, void *, size_t, int *);
-static ssize_t write_fd(int, void *, size_t, int);
+NOEXPORT ssize_t read_fd(int, void *, size_t, int *);
+NOEXPORT ssize_t write_fd(int, void *, size_t, int);
 
 int num_processes=0;
 static int *ipc_socket, *busy;
-#endif /* USE_PTHREAD */
+#endif /* USE_LIBWRAP_POOL */
 
-void libwrap_init() {
-#ifdef USE_PTHREAD
+int libwrap_init() {
+#ifdef USE_LIBWRAP_POOL
     int i, j, rfd, result;
     char servname[SERVNAME_LEN];
     static int initialized=0;
     SERVICE_OPTIONS *opt;
 
     if(initialized) /* during startup or previous configuration file reload */
-        return;
+        return 0;
     for(opt=service_options.next; opt; opt=opt->next)
         if(opt->option.libwrap) /* libwrap is enabled for this service */
             break;
     if(!opt) /* disabled for all sections or inetd mode (no sections) */
-        return;
+        return 0;
 
     num_processes=LIBWRAP_CLIENTS;
     ipc_socket=str_alloc(2*num_processes*sizeof(int));
     busy=str_alloc(num_processes*sizeof(int));
     for(i=0; i<num_processes; ++i) { /* spawn a child */
         if(s_socketpair(AF_UNIX, SOCK_STREAM, 0, ipc_socket+2*i, 0, "libwrap_init"))
-            die(1);
+            return 1;
         switch(fork()) {
         case -1:    /* error */
             ioerror("fork");
-            die(1);
+            return 1;
         case  0:    /* child */
             drop_privileges(0); /* libwrap processes are not chrooted */
             close(0); /* stdin */
@@ -102,17 +107,18 @@ void libwrap_init() {
         }
     }
     initialized=1;
-#endif /* USE_PTHREAD */
+#endif /* USE_LIBWRAP_POOL */
+    return 0;
 }
 
 void libwrap_auth(CLI *c, char *accepted_address) {
     int result=0; /* deny by default */
-#ifdef USE_PTHREAD
+#ifdef USE_LIBWRAP_POOL
     static volatile int num_busy=0, roundrobin=0;
     int retval, my_process;
     static pthread_mutex_t mutex=PTHREAD_MUTEX_INITIALIZER;
     static pthread_cond_t cond=PTHREAD_COND_INITIALIZER;
-#endif /* USE_PTHREAD */
+#endif /* USE_LIBWRAP_POOL */
 
     if(!c->opt->option.libwrap) /* libwrap is disabled for this service */
         return; /* allow connection */
@@ -122,7 +128,7 @@ void libwrap_auth(CLI *c, char *accepted_address) {
         return;
     }
 #endif
-#ifdef USE_PTHREAD
+#ifdef USE_LIBWRAP_POOL
     if(num_processes) {
         s_log(LOG_DEBUG, "Waiting for a libwrap process");
 
@@ -155,7 +161,7 @@ void libwrap_auth(CLI *c, char *accepted_address) {
         s_log(LOG_DEBUG, "Acquired libwrap process #%d", my_process);
         write_fd(ipc_socket[2*my_process], c->opt->servname,
             strlen(c->opt->servname)+1, c->local_rfd.fd);
-        read_blocking(c, ipc_socket[2*my_process],
+        s_read(c, ipc_socket[2*my_process],
             (u8 *)&result, sizeof result);
         s_log(LOG_DEBUG, "Releasing libwrap process #%d", my_process);
 
@@ -167,13 +173,11 @@ void libwrap_auth(CLI *c, char *accepted_address) {
         }
         busy[my_process]=0; /* mark the child process as free */
         --num_busy; /* the child process has been released */
-        if(num_busy==num_processes-1) { /* need to wake up a thread */
-            retval=pthread_cond_signal(&cond); /* signal waiting threads */
-            if(retval) {
-                errno=retval;
-                ioerror("pthread_cond_signal");
-                longjmp(c->err, 1);
-            }
+        retval=pthread_cond_signal(&cond); /* signal a waiting thread */
+        if(retval) {
+            errno=retval;
+            ioerror("pthread_cond_signal");
+            longjmp(c->err, 1);
         }
         retval=pthread_mutex_unlock(&mutex);
         if(retval) {
@@ -184,23 +188,23 @@ void libwrap_auth(CLI *c, char *accepted_address) {
 
         s_log(LOG_DEBUG, "Released libwrap process #%d", my_process);
     } else
-#endif /* USE_PTHREAD */
+#endif /* USE_LIBWRAP_POOL */
     { /* use original, synchronous libwrap calls */
         enter_critical_section(CRIT_LIBWRAP);
         result=check(c->opt->servname, c->local_rfd.fd);
         leave_critical_section(CRIT_LIBWRAP);
     }
     if(!result) {
-        s_log(LOG_WARNING, "Service %s REFUSED by libwrap from %s",
+        s_log(LOG_WARNING, "Service [%s] REFUSED by libwrap from %s",
             c->opt->servname, accepted_address);
         s_log(LOG_DEBUG, "See hosts_access(5) manual for details");
         longjmp(c->err, 1);
     }
-    s_log(LOG_DEBUG, "Service %s permitted by libwrap from %s",
+    s_log(LOG_DEBUG, "Service [%s] permitted by libwrap from %s",
         c->opt->servname, accepted_address);
 }
 
-static int check(char *name, int fd) {
+NOEXPORT int check(char *name, int fd) {
     struct request_info request;
 
     request_init(&request, RQ_DAEMON, name, RQ_FILE, fd, 0);
@@ -208,9 +212,9 @@ static int check(char *name, int fd) {
     return hosts_access(&request);
 }
 
-#ifdef USE_PTHREAD
+#ifdef USE_LIBWRAP_POOL
 
-static ssize_t read_fd(int fd, void *ptr, size_t nbytes, int *recvfd) {
+NOEXPORT ssize_t read_fd(int fd, void *ptr, size_t nbytes, int *recvfd) {
     struct msghdr msg;
     struct iovec iov[1];
     ssize_t n;
@@ -265,7 +269,7 @@ static ssize_t read_fd(int fd, void *ptr, size_t nbytes, int *recvfd) {
     return n;
 }
 
-static ssize_t write_fd(int fd, void *ptr, size_t nbytes, int sendfd) {
+NOEXPORT ssize_t write_fd(int fd, void *ptr, size_t nbytes, int sendfd) {
     struct msghdr msg;
     struct iovec iov[1];
 
@@ -300,7 +304,7 @@ static ssize_t write_fd(int fd, void *ptr, size_t nbytes, int sendfd) {
     return sendmsg(fd, &msg, 0);
 }
 
-#endif /* USE_PTHREAD */
+#endif /* USE_LIBWRAP_POOL */
 
 #endif /* USE_LIBWRAP */
 
